/*
 * Copyright (C) 1997-2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */


#include "server.h"

extern struct misc_meta_data *mmd;
extern struct dbtool dblist[];

static FILE *audio_file = NULL;
static char supported_audio_formats[MAXLINE];

#if 1
	int mp3_init(void *);
#endif

#ifdef HAVE_OGGVORBIS
	int ogg_init(void *);
#endif

static struct audio_format afl[] = { /* audio format list */
#if 1
	{
		.name = "mp3",
		.init = mp3_init,
		.sw = NULL,
		.sw_pid = 0
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.init = ogg_init,
		.sw = NULL,
		.sw_pid = 0
	},
#endif
	{
		.name = NULL,
	}
};

const char *audio_format_name(int i)
{
	return i >= 0?  afl[i].name : "(none)";
}

char *afs_stream_writer_cmd(int i)
{
	return i >=0? afl[i].open_cmd : "/bin/false";
}


char *afs_supported_audio_formats(void)
{
	return supported_audio_formats;
}

static FILE *afs_open_stream_writer(int format_num)
{
	struct audio_format *af;
	long flags;
	int file_no;

	if (format_num < 0)
		return NULL;
	af = afl + format_num;

	clog(INFO, "%s: opening %s stream_writer (%s)\n", __func__, af->name,
		af->open_cmd);
	af->sw = popen_write_cmdline_pid(&af->sw_pid, af->open_cmd);
	if (!af->sw) {
		clog(ERROR, "%s: failed to open %s stream writer\n", __func__,
			af->name);
		return NULL;
	}
	file_no = fileno(af->sw);
	clog(DEBUG, "%s: %s stream writer pid: %d, fd: %d\n", __func__, 
		af->name, af->sw_pid, file_no);
	/* set non-blocking io */
	flags = fcntl(file_no, F_GETFL) | O_NONBLOCK;
	fcntl(file_no, F_SETFL, flags);
	return af->sw;
}

static int setup_stream_command(struct gengetopt_args_info *conf, int num)
{
	int i, ret = 0;
	const char *n = afl[num].name;
	int g = conf->stream_write_cmd_given;

	if (!g)
		goto out;
	clog(DEBUG, "%s: %d stream write command%s given, checking for %s\n", 
		__func__, g, g > 1? "s" : "", n);
	for (i = g - 1; i >= 0; i--) {
		char *c = conf->stream_write_cmd_arg[i];
		size_t len = strlen(n);
		if (strlen(c) <= len || strncmp(c, n, len))
			continue;
		if (*(c + len) != ':') {
			clog(WARNING, "%s: invalid stream_write_cmd\n", __func__);
			return -1;
		}
		afl[num].open_cmd = c + len + 1;
		ret = 1;
		break;
		
	}
out:
	clog(INFO, "%s: using \"%s\" for files of type %s%s\n", 
		__func__, afl[num].open_cmd, afl[num].name,
		ret == 1? "" : " (default)");
	return ret;
}

int afs_handle_sighup(struct gengetopt_args_info *conf)
{
	int i, ret;
	for (i = 0; afl[i].name; i++)
		if ((ret = setup_stream_command(conf, i)) < 0)
			return ret;
	return 1;
}

int afs_init(struct gengetopt_args_info *conf)
{
	int i, ret;

	clog(INFO, "%s: checking audio formats\n", __func__);
	*supported_audio_formats = '\0';
	for (i = 0; afl[i].name; i++) {
		clog(DEBUG, "%s: supported audio format: %s\n", __func__, 
			afl[i].name);
		if (afl[i].init(&afl[i]) < 0) {
			clog(ERROR, "%s: init %s failed, disabling "
				"%s support\n", __func__, afl[i].name, 
				afl[i].name);
			afl[i].get_file_info = NULL;
			continue;
		}
		if ((ret = setup_stream_command(conf, i)) < 0)
			return ret;
		sprintf(supported_audio_formats, "%s%s ", 
			*supported_audio_formats? supported_audio_formats: "",
			 afl[i].name);
//		afl[i].sw_pid = 0;
//		afl[i].sw = NULL;
	}
	if (!i) {
		clog(EMERG, "%s: no audio formats supported\n");
		return -1;
	}
	clog(DEBUG, "%s: supported audio formats: %s\n", __func__, 
			supported_audio_formats);
	
	return i;
}

static int get_file_info(int i)
{
	int ret = afl[i].get_file_info(audio_file, mmd->audio_file_info,
		&mmd->chunks_total, &mmd->seconds_total);
	if (ret > 0)
		clog(INFO, "%s: valid %s file\n", __func__, afl[i].name);
	else
		clog(DEBUG, "%s: %s format not recognized\n", __func__, 
			afl[i].name);
	return ret;
}

static int guess_audio_format(const char *name)
{

	int i, len1 = strlen(name), len2;

	for (i = 0; afl[i].name; i++) {
		len2 = strlen(afl[i].name);
		if (len1 < len2)
			continue;
		if (!strncasecmp(name + (len1 - len2), afl[i].name, len2)) {
			clog(DEBUG, "%s: might be %s\n", __func__,
				afl[i].name);
			return i;
		}
	}
	return -1;
}

static int get_audio_format(int omit)
{
	int i;

	for (i = 0; afl[i].name; i++) {
		if (i == omit || !afl[i].get_file_info)
			continue;
		rewind(audio_file);
		if (get_file_info(i) > 0)
			return i;
		rewind(audio_file);
	}
	return -1;
}

/*
 * upddate shared mem
 */
static int update_mmd(void)
{
	int i;
	struct stat file_status;

	i = guess_audio_format(mmd->filename);
	if (i < 0 || get_file_info(i) < 0)
		i = get_audio_format(i);
	if (i < 0) {
		clog(WARNING, "%s: unknown audio format for %s\n", __func__,
			mmd->filename);
		return -1;
	}
	mmd->audio_format = i;
	mmd->chunks_sent = 0;
	if (fstat(fileno(audio_file), &file_status) == -1) {
		clog(ERROR, "%s: fstat error (%s)\n", __func__, strerror(errno));
		fclose(audio_file);
		return -1;
	}
	mmd->size = file_status.st_size;
	mmd->mtime = file_status.st_mtime;
	clog(NOTICE, "%s: next audio file: %s\n", __func__, mmd->filename);
	return 1;
}

static void get_song(void)
{
	char **sl;
	int i;

	sl = dblist[mmd->dbt_num].get_song_list(10);
	if (!sl)
		goto err_out;
	for (i = 0; sl[i]; i++) {
		clog(INFO, "%s: trying %s\n", __func__, sl[i]); 
		if (strlen(sl[i]) >= _POSIX_PATH_MAX)
			continue;
		audio_file = fopen(sl[i], "r");
		if (!audio_file)
			continue;
		strcpy(mmd->filename, sl[i]);
		if (update_mmd() < 0)
			continue;
		mmd->num_played++;
		dblist[mmd->dbt_num].update_song(sl[i]);
		clog(DEBUG, "%s: success\n", __func__); 
		mmd->afs_status_flags &= (~AFS_NEXT);
		goto free;
	}
err_out:
	mmd->afs_status_flags = AFS_NEXT;
	clog(ERROR, "%s: no valid files found\n", __func__);
free:
	if (sl) {
		for (i = 0; sl[i]; i++)
			free(sl[i]);
		free(sl);
	}
}

/*
 * called from main when playing, just before main calls select
 */
FILE *afs_preselect(void)
{
	struct audio_format *af;

	if (!audio_file) {
		clog(DEBUG, "%s: no audio file\n", __func__);
		get_song();
	}
	if (!(mmd->afs_status_flags & AFS_PLAYING))
		return NULL;
	af = afl + mmd->audio_format;
	if (!af->sw) {
		clog(DEBUG, "%s: open stream\n", __func__); 
		af->sw = afs_open_stream_writer(mmd->audio_format);
	}
	return af->sw;
}

/*
 * Status is playing and stream writer is ready for writing.
 */
int afs_send_chunk()
{
	int ret;
	struct audio_format *af = &afl[mmd->audio_format];

	if (mmd->audio_format < 0 || !audio_file) {
		clog(CRIT, "%s: BUG\n", __func__);
		return -1;
	}
	ret = af->send_chunk(&af->sw, mmd->audio_file_info, 
		&mmd->chunks_sent, mmd->afs_status_flags, &af->sw_pid);
	mmd->afs_status_flags &= ~(AFS_NEXT | AFS_REPOS);
	if (ret <= 0)
		audio_file = NULL;
	else
		mmd->afs_status_flags |= AFS_PLAYING;
	if (ret < 0 || (mmd->afs_status_flags & AFS_NOMORE))
		mmd->afs_status_flags = AFS_NEXT;
	return ret;
}
